import platform
from pathlib import Path

from click import get_app_dir
from xdg_base_dirs import xdg_config_home, xdg_data_home


def qutebrowser_exe() -> str:
    macos_app = "/Applications/qutebrowser.app/Contents/MacOS/qutebrowser"
    if platform == "darwin" and Path(macos_app).exists():
        return macos_app
    else:
        return "qutebrowser"


def default_qbpm_application_dir() -> Path:
    path = xdg_data_home() / "applications" / "qbpm"
    path.mkdir(parents=True, exist_ok=True)
    return path


def default_profile_dir() -> Path:
    path = xdg_data_home() / "qutebrowser-profiles"
    path.mkdir(parents=True, exist_ok=True)
    return path


def qutebrowser_data_dir() -> Path:
    if platform.system() == "Linux":
        return xdg_data_home() / "qutebrowser"
    # TODO confirm this works on windows
    return Path(get_app_dir("qutebrowser", roaming=True))


def qutebrowser_config_dirs() -> list[Path]:
    # deduplicate while maintaining order
    return list(
        dict.fromkeys(
            [
                Path(get_app_dir("qutebrowser", roaming=True)),
                xdg_config_home() / "qutebrowser",
                Path.home() / ".qutebrowser",
            ]
        )
    )



from pathlib import Path
from typing import Optional

from .log import error
from .paths import qutebrowser_exe

try:
    from qbpm.version import version as __version__  # type: ignore
except ImportError:
    __version__ = "unknown"


class Profile:
    name: str
    profile_dir: Path
    root: Path

    def __init__(self, name: str, profile_dir: Path) -> None:
        self.name = name
        self.profile_dir = profile_dir
        self.root = self.profile_dir / name

    def check(self) -> Optional["Profile"]:
        if "/" in self.name:
            error("profile name cannot contain slashes")
            return None
        return self

    def exists(self) -> bool:
        return self.root.exists() and self.root.is_dir()

    def cmdline(self) -> list[str]:
        return [
            qutebrowser_exe(),
            "-B",
            str(self.root),
            "--qt-arg",
            "name",
            self.name,
            "--desktop-file-name",
            self.name,
        ]



import subprocess
from pathlib import Path

from . import Profile
from .launch import launch_qutebrowser
from .log import error
from .menus import find_menu


def choose_profile(
    profile_dir: Path, menu: str | None, foreground: bool, qb_args: tuple[str, ...]
) -> bool:
    dmenu = find_menu(menu)
    if not dmenu:
        return False

    real_profiles = {profile.name for profile in profile_dir.iterdir()}
    if len(real_profiles) == 0:
        error("no profiles")
        return False
    profiles = [*real_profiles, "qutebrowser"]
    command = dmenu.command(sorted(profiles), "qutebrowser", " ".join(qb_args))
    selection_cmd = subprocess.run(
        command,
        text=True,
        input="\n".join(sorted(profiles)),
        stdout=subprocess.PIPE,
        stderr=None,
        check=False,
    )
    out = selection_cmd.stdout
    selection = out.rstrip("\n")

    if selection == "qutebrowser" and "qutebrowser" not in real_profiles:
        return launch_qutebrowser(None, foreground, qb_args)
    elif selection:
        profile = Profile(selection, profile_dir)
        return launch_qutebrowser(profile, foreground, qb_args)
    else:
        error("no profile selected")
        return False



import textwrap
from pathlib import Path

from . import Profile
from .paths import default_qbpm_application_dir

MIME_TYPES = [
    "text/html",
    "text/xml",
    "application/xhtml+xml",
    "application/xml",
    "application/rdf+xml",
    "image/gif",
    "image/jpeg",
    "image/png",
    "x-scheme-handler/http",
    "x-scheme-handler/https",
    "x-scheme-handler/qute",
]


# TODO expose application_dir through config
def create_desktop_file(profile: Profile, application_dir: Path | None = None) -> None:
    text = textwrap.dedent(f"""\
        [Desktop Entry]
        Name={profile.name} (qutebrowser profile)
        StartupWMClass=qutebrowser
        GenericName={profile.name}
        Icon=qutebrowser
        Type=Application
        Categories=Network;WebBrowser;
        Exec={" ".join([*profile.cmdline(), "--untrusted-args", "%u"])}
        Terminal=false
        StartupNotify=true
        MimeType={";".join(MIME_TYPES)};
        Keywords=Browser
        Actions=new-window;preferences;

        [Desktop Action new-window]
        Name=New Window
        Exec={" ".join(profile.cmdline())}

        [Desktop Action preferences]
        Name=Preferences
        Exec={" ".join([*profile.cmdline(), '"qute://settings"'])}
    """)
    application_dir = application_dir or default_qbpm_application_dir()
    (application_dir / f"{profile.name}.desktop").write_text(text)



import platform
import shlex
import sys
from collections.abc import Iterator
from dataclasses import dataclass, replace
from os import environ
from pathlib import Path
from shutil import which

from .log import error, or_phrase


@dataclass(frozen=True)
class Dmenu:
    menu_command: list[str]

    def name(self) -> str:
        return self.menu_command[0]

    def installed(self) -> bool:
        return which(self.name()) is not None

    def command(self, _profiles: list[str], prompt: str, qb_args: str) -> list[str]:
        return [arg.format(prompt=prompt, qb_args=qb_args) for arg in self.menu_command]


class ApplescriptMenu:
    @classmethod
    def name(cls) -> str:
        return "applescript"

    @classmethod
    def installed(cls) -> bool:
        return platform.system() == "Darwin"

    @classmethod
    def command(cls, profiles: list[str], _prompt: str, qb_args: str) -> list[str]:
        profile_list = '", "'.join(profiles)
        return [
            "osascript",
            "-e",
            f"""set profiles to {{"{profile_list}"}}
set profile to choose from list profiles with prompt "qutebrowser: {qb_args}" default items {{item 1 of profiles}}
item 1 of profile""",
        ]


def find_menu(menu: str | None) -> Dmenu | ApplescriptMenu | None:
    menus = list(supported_menus())
    if not menu:
        found = next(filter(lambda m: m.installed(), menus), None)
        if not found:
            error(
                "no menu program found, use --menu to provide a dmenu-compatible menu or install one of "
                + or_phrase([m.name() for m in menus if isinstance(m, Dmenu)])
            )
        return found
    dmenu = custom_dmenu(menu)
    if not dmenu.installed():
        error(f"{dmenu.name()} not found")
        return None
    return dmenu


def custom_dmenu(command: str) -> Dmenu:
    split = shlex.split(command)
    if len(split) == 1 or not split[1]:
        name = Path(command).name
        for menu in supported_menus():
            if isinstance(menu, Dmenu) and menu.name() == name:
                return (
                    menu
                    if name == command
                    else replace(menu, menu_command=[command, *menu.menu_command[1::]])
                )
    return Dmenu(split)


def supported_menus() -> Iterator[Dmenu | ApplescriptMenu]:
    if ApplescriptMenu.installed():
        yield ApplescriptMenu()
    if environ.get("WAYLAND_DISPLAY"):
        yield from [
            # default window is too narrow for a long prompt
            Dmenu(["fuzzel", "--dmenu"]),
            Dmenu(["walker", "--dmenu", "--placeholder", "{prompt} {qb_args}"]),
            Dmenu(["wofi", "--dmenu", "--prompt", "{prompt} {qb_args}"]),
            Dmenu(["tofi", "--prompt-text", "{prompt}> "]),
            Dmenu(["wmenu", "-p", "{prompt}"]),
            Dmenu(["dmenu-wl", "--prompt", "{prompt}"]),
        ]
    if environ.get("DISPLAY"):
        yield from [
            Dmenu(
                [
                    "rofi",
                    "-dmenu",
                    "-no-custom",
                    "-p",
                    "{prompt}",
                    "-mesg",
                    "{qb_args}",
                ]
            ),
            Dmenu(["dmenu", "-p", "{prompt}"]),
        ]
    if sys.stdin.isatty():
        if environ.get("TMUX"):
            yield Dmenu(["fzf-tmux", "--prompt", "{prompt}> "])
        yield Dmenu(["fzf", "--prompt", "{prompt}> "])



from functools import partial
from pathlib import Path
from sys import platform

from . import Profile
from .desktop import create_desktop_file
from .log import error, or_phrase
from .paths import qutebrowser_config_dirs

MIME_TYPES = [
    "text/html",
    "text/xml",
    "application/xhtml+xml",
    "application/xml",
    "application/rdf+xml",
    "image/gif",
    "image/jpeg",
    "image/png",
    "x-scheme-handler/http",
    "x-scheme-handler/https",
    "x-scheme-handler/qute",
]


def create_profile(profile: Profile, overwrite: bool = False) -> bool:
    if not profile.check():
        return False

    if not overwrite and profile.root.exists():
        error(f"{profile.root} already exists")
        return False

    config_dir = profile.root / "config"
    config_dir.mkdir(parents=True, exist_ok=overwrite)
    print(profile.root)
    return True


def create_config(
    profile: Profile,
    qb_config_dir: Path,
    home_page: str | None = None,
    overwrite: bool = False,
) -> None:
    user_config = profile.root / "config" / "config.py"
    with user_config.open(mode="w" if overwrite else "x") as dest_config:
        out = partial(print, file=dest_config)
        out("config.load_autoconfig()")
        title_prefix = "{perc}{current_title}{title_sep}"
        out(f"c.window.title_format = '{title_prefix} qutebrowser ({profile.name})'")
        if home_page:
            out(f"c.url.start_pages = ['{home_page}']")
        out(f"config.source(r'{qb_config_dir / 'config.py'}')")


def exists(profile: Profile) -> bool:
    if profile.root.exists() and not profile.root.is_dir():
        error(f"{profile.root} is not a directory")
        return False
    if not profile.root.exists():
        error(f"{profile.root} does not exist")
        return False
    return True


def new_profile(
    profile: Profile,
    qb_config_dir: Path | None,
    home_page: str | None = None,
    desktop_file: bool | None = None,
    overwrite: bool = False,
) -> bool:
    qb_config_dir = find_qutebrowser_config_dir(qb_config_dir)
    if not qb_config_dir:
        return False
    if create_profile(profile, overwrite):
        create_config(profile, qb_config_dir, home_page, overwrite)
        if desktop_file is True or (desktop_file is not False and platform == "linux"):
            create_desktop_file(profile)
        return True
    return False


def find_qutebrowser_config_dir(qb_config_dir: Path | None) -> Path | None:
    config_file = "config.py"
    dirs = (
        [qb_config_dir, qb_config_dir / "config"]
        if qb_config_dir
        else qutebrowser_config_dirs()
    )
    for config_dir in dirs:
        if (config_dir / config_file).exists():
            return config_dir.absolute()
    error(f"could not find {config_file} in {or_phrase(dirs)}")
    return None



import logging
import sys
from collections.abc import Callable
from dataclasses import dataclass
from functools import wraps
from pathlib import Path
from typing import Any, NoReturn, TypeVar

import click

from . import Profile, operations, profiles
from .choose import choose_profile
from .launch import launch_qutebrowser
from .log import error, or_phrase
from .menus import supported_menus
from .paths import default_profile_dir, qutebrowser_data_dir

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@dataclass
class Context:
    profile_dir: Path


@dataclass
class CreatorOptions:
    qb_config_dir: Path | None
    launch: bool
    foreground: bool
    desktop_file: bool
    overwrite: bool


T = TypeVar("T")


def creator_options(orig: Callable[..., T]) -> Callable[..., T]:
    @wraps(orig)
    def command(
        qb_config_dir: Path | None,
        launch: bool,
        foreground: bool,
        desktop_file: bool,
        overwrite: bool,
        *args: Any,  # noqa: ANN401
        **kwargs: Any,  # noqa: ANN401
    ) -> T:
        return orig(
            *args,
            c_opts=CreatorOptions(
                qb_config_dir, launch, foreground, desktop_file, overwrite
            ),
            **kwargs,
        )

    for opt in reversed(
        [
            click.option(
                "-C",
                "--qutebrowser-config-dir",
                "qb_config_dir",
                type=click.Path(file_okay=False, readable=True, path_type=Path),
                help="Location of the qutebrowser config to inherit from.",
            ),
            click.option("-l", "--launch", is_flag=True, help="Launch the profile."),
            click.option(
                "-f",
                "--foreground",
                is_flag=True,
                help="If --launch is set, run qutebrowser in the foreground.",
            ),
            click.option(
                "--no-desktop-file",
                "desktop_file",
                default=True,
                is_flag=True,
                flag_value=False,
                help="Do not generate an XDG desktop entry for the profile.",
            ),
            click.option(
                "--overwrite",
                is_flag=True,
                help="Replace the current profile configuration if it exists.",
            ),
        ]
    ):
        command = opt(command)
    return command


class LowerCaseFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        record.levelname = record.levelname.lower()
        return super().format(record)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option()
@click.option(
    "-P",
    "--profile-dir",
    type=click.Path(file_okay=False, writable=True, path_type=Path),
    envvar="QBPM_PROFILE_DIR",
    show_envvar=True,
    default=None,
    help="Location to store qutebrowser profiles.",
)
@click.option(
    "-l",
    "--log-level",
    default="error",
    type=click.Choice(["debug", "info", "error"], case_sensitive=False),
)
@click.pass_context
def main(ctx: click.Context, profile_dir: Path | None, log_level: str) -> None:
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level.upper())
    handler = logging.StreamHandler()
    handler.setFormatter(LowerCaseFormatter("{levelname}: {message}", style="{"))
    root_logger.addHandler(handler)
    ctx.obj = Context(profile_dir or default_profile_dir())


@main.command()
@click.argument("profile_name")
@click.argument("home_page", required=False)
@creator_options
@click.pass_obj
def new(
    context: Context,
    profile_name: str,
    home_page: str | None,
    c_opts: CreatorOptions,
) -> None:
    """Create a new profile."""
    profile = Profile(profile_name, **vars(context))
    exit_with(
        profiles.new_profile(
            profile,
            c_opts.qb_config_dir,
            home_page,
            c_opts.desktop_file,
            c_opts.overwrite,
        )
        and ((not c_opts.launch) or launch_qutebrowser(profile, c_opts.foreground))
    )


@main.command()
@click.argument("session")
@click.argument("profile_name", required=False)
@creator_options
@click.pass_obj
def from_session(
    context: Context,
    session: str,
    profile_name: str | None,
    c_opts: CreatorOptions,
) -> None:
    """Create a new profile from a saved qutebrowser session.

    SESSION may be the name of a session in the global qutebrowser profile
    or a path to a session yaml file.
    """
    profile, session_path = session_info(session, profile_name, context)
    exit_with(
        operations.from_session(
            profile,
            session_path,
            c_opts.qb_config_dir,
            c_opts.desktop_file,
            c_opts.overwrite,
        )
        and ((not c_opts.launch) or launch_qutebrowser(profile, c_opts.foreground))
    )


@main.command("launch", context_settings={"ignore_unknown_options": True})
@click.argument("profile_name")
@click.argument("qb_args", nargs=-1, type=click.UNPROCESSED)
@click.option(
    "-f", "--foreground", is_flag=True, help="Run qutebrowser in the foreground."
)
@click.pass_obj
def launch_profile(
    context: Context, profile_name: str, foreground: bool, qb_args: tuple[str, ...]
) -> None:
    """Launch qutebrowser with a specific profile.

    All QB_ARGS are passed on to qutebrowser."""
    profile = Profile(profile_name, **vars(context))
    exit_with(launch_qutebrowser(profile, foreground, qb_args))


@main.command(context_settings={"ignore_unknown_options": True})
@click.argument("qb_args", nargs=-1, type=click.UNPROCESSED)
@click.option(
    "-m",
    "--menu",
    metavar="COMMAND",
    help="A dmenu-compatible command or one of the following supported menus: "
    + ", ".join([menu.name() for menu in supported_menus()]),
)
@click.option(
    "-f", "--foreground", is_flag=True, help="Run qutebrowser in the foreground."
)
@click.pass_obj
def choose(
    context: Context, menu: str | None, foreground: bool, qb_args: tuple[str, ...]
) -> None:
    """Choose a profile to launch.

    Support is built in for many X and Wayland launchers, as well as applescript dialogs.
    All QB_ARGS are passed on to qutebrowser.
    """
    exit_with(choose_profile(context.profile_dir, menu, foreground, qb_args))


@main.command()
@click.argument("profile_name")
@click.pass_obj
def edit(context: Context, profile_name: str) -> None:
    """Edit a profile's config.py."""
    profile = Profile(profile_name, **vars(context))
    if not profile.exists():
        error(f"profile {profile.name} not found at {profile.root}")
        sys.exit(1)
    click.edit(filename=str(profile.root / "config" / "config.py"))


@main.command(name="list")
@click.pass_obj
def list_(context: Context) -> None:
    """List existing profiles."""
    for profile in sorted(context.profile_dir.iterdir()):
        print(profile.name)


@main.command()
@click.argument("profile_name")
@click.pass_obj
def desktop(
    context: Context,
    profile_name: str,
) -> None:
    """Create an XDG desktop entry for an existing profile."""
    profile = Profile(profile_name, **vars(context))
    exit_with(operations.desktop(profile))


def session_info(
    session: str, profile_name: str | None, context: Context
) -> tuple[Profile, Path]:
    user_session_dir = qutebrowser_data_dir() / "sessions"
    session_paths = []
    if "/" not in session:
        session_paths.append(user_session_dir / (session + ".yml"))
    session_paths.append(Path(session))
    session_path = next(filter(lambda path: path.is_file(), session_paths), None)

    if not session_path:
        tried = or_phrase([str(p.resolve()) for p in session_paths])
        error(f"could not find session file at {tried}")
        sys.exit(1)

    return (Profile(profile_name or session_path.stem, **vars(context)), session_path)


def exit_with(result: bool) -> NoReturn:
    sys.exit(0 if result else 1)



import shutil
import subprocess

from . import Profile
from .log import error
from .paths import qutebrowser_exe


def launch_qutebrowser(
    profile: Profile | None, foreground: bool, qb_args: tuple[str, ...] = ()
) -> bool:
    qb = profile.cmdline() if profile else [qutebrowser_exe()]
    return launch(foreground, [*qb, *qb_args])


def launch(foreground: bool, args: list[str]) -> bool:
    if not shutil.which(args[0]):
        error("qutebrowser is not installed")
        return False

    if foreground:
        return subprocess.run(args, check=False).returncode == 0
    else:
        p = subprocess.Popen(args, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)
        try:
            # give qb a chance to validate input before returning to shell
            stdout, stderr = p.communicate(timeout=0.1)
            print(stderr.decode(errors="ignore"), end="")
        except subprocess.TimeoutExpired:
            pass

    return True



import shutil
from pathlib import Path

from . import Profile, profiles
from .desktop import create_desktop_file
from .log import error


def from_session(
    profile: Profile,
    session_path: Path,
    qb_config_dir: Path | None,
    desktop_file: bool = True,
    overwrite: bool = False,
) -> bool:
    if not profiles.new_profile(profile, qb_config_dir, None, desktop_file, overwrite):
        return False

    session_dir = profile.root / "data" / "sessions"
    session_dir.mkdir(parents=True, exist_ok=overwrite)
    shutil.copy(session_path, session_dir / "_autosave.yml")

    return True


def desktop(profile: Profile) -> bool:
    exists = profile.exists()
    if exists:
        create_desktop_file(profile)
    else:
        error(f"profile {profile.name} not found at {profile.root}")
    return exists



import logging


def info(msg: str) -> None:
    logging.info(msg)


def error(msg: str) -> None:
    logging.error(msg)


def or_phrase(items: list) -> str:
    strings = list(map(str, items))
    size = len(strings)
    if size == 0:
        return "[]"
    elif size == 1:
        return strings[0]
    elif size == 2:  # noqa: PLR2004
        return " or ".join(strings)
    else:
        return ", or ".join([", ".join(strings[0:-1]), strings[-1]])
